/*************************************************************************
 *
 *   Used with ICCARM and AARM.
 *
 *    (c) Copyright IAR Systems 2012
 *
 *    File name   : Flash.c
 *    Description : FlashLoader  internal flash loader
 *
 *    History :
 *    1. Date        : July, 2010
 *       Author      : Stoyan Choynev
 *       Description : Initial revision
 *    2. Date        : 10 September, 2012
 *       Author      : Stanimir Bonev
 *       Description : port for KMxx series
 *
 *    $Revision: 16 $
 **************************************************************************/

#include <intrinsics.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <freescale/MKM34Z5.h>

#include "flash_loader.h"       // The flash loader framework API declarations.
#include "flash_loader_extra.h"

/*
** define flash command codes
*/
#define ERPS      0x09
#define WRP       0x06
#define ERALL     0x44

typedef uint32_t flashunit;

/** default settings **/

/** external functions **/

/** external data **/

/** internal functions **/
static uint32_t CommandLaunch(void);
static void DisableWDT(void);
__no_init static uint32_t user_supply_prot;

/** public data **/
/** private data **/
__no_init static volatile uint32_t flashBase;

/** public functions **/

/*************************************************************************
 * Function Name: FlashInit
 * Parameters: Flash Base Address
 *
 * Return:  0 - Init Successful
 *          1 - Init Fail
 * Description: Init flash and build layout.
 *
 *************************************************************************/
#if USE_ARGC_ARGV
uint32_t FlashInit(void *base_of_flash, uint32_t image_size,
                   uint32_t link_address, uint32_t flags,
                   int argc, char const *argv[])
#else
uint32_t FlashInit(void *base_of_flash, uint32_t image_size,
                   uint32_t link_address, uint32_t flags)
#endif /* USE_ARGC_ARGV */
{
  DisableWDT();
#if USE_ARGC_ARGV
  user_supply_prot = 0;
  for(int i = 0; i < argc; i++)
  {
    /*if there is a partition arg*/
    if(strcmp("--enable_config_write", argv[i]) == 0)
    {
      user_supply_prot = 1;
    }
  }
#else
  user_supply_prot = 1;
#endif
  /* wait until command complete */
  while(!(FTFA_FSTAT & FTFA_FSTAT_CCIF_MASK));
  /* Clear status */
  FTFA_FSTAT = FTFA_FSTAT_ACCERR_MASK | FTFA_FSTAT_FPVIOL_MASK;

  if (FLAG_ERASE_ONLY & flags)
  {
    uint32_t status;
    // chip erase
    FTFA_FCCOB0 = ERALL;
    status = CommandLaunch();

    /* Unlock device wneh erese sector 0*/
    if(!user_supply_prot)
    {
      union
      {
        uint32_t  word;
        uint8_t   byte[4];
      } dest;

      dest.word   = 0x40C;
      FTFA_FCCOB0 = WRP;
      FTFA_FCCOB1 = dest.byte[2];
      FTFA_FCCOB2 = dest.byte[1];
      FTFA_FCCOB3 = dest.byte[0];
      FTFA_FCCOB4 = 0xFF;         /* 0x40F */
      FTFA_FCCOB5 = 0xFF;         /* 0x40E */
      FTFA_FCCOB6 = 0xFF;         /* 0x40D */
      FTFA_FCCOB7 = 0xFE;         /* 0x40C */

      if(   (RESULT_OK != CommandLaunch())
         || (RESULT_OK != status)) return RESULT_ERROR;
    }
    return RESULT_ERASE_DONE;
  }

  return (RESULT_OK);
}

/*************************************************************************
 * Function Name: FlashWrite
 * Parameters: block base address, offet in block, data size, ram buffer
 *             pointer
 * Return:  0 - Write Successful
 *          1 - Write Fail
 * Description. Writes data to Flash
 *************************************************************************/
uint32_t FlashWrite(void *block_start,
                    uint32_t offset_into_block,
                    uint32_t count,
                    char const *buffer)
{
uint32_t size;

union
{
  uint32_t  word;
  uint8_t   byte[4];
} dest;

  /*Set Write command*/
  while(!(FTFA_FSTAT & FTFA_FSTAT_CCIF_MASK));

  for(size = 0, dest.word = (uint32_t)block_start + offset_into_block;
      size < count; size += 4, dest.word += 4, buffer += 4)
  {
    if(dest.word == 0x40c && !user_supply_prot)
      continue;

    /*Set destination address and command*/
    FTFA_FCCOB0 = WRP;
    FTFA_FCCOB1 = dest.byte[2];
    FTFA_FCCOB2 = dest.byte[1];
    FTFA_FCCOB3 = dest.byte[0];
    /*copy data*/
    FTFA_FCCOB4 = buffer[3]; /* +3 */
    FTFA_FCCOB5 = buffer[2]; /* +2 */
    FTFA_FCCOB6 = buffer[1]; /* +1 */
    FTFA_FCCOB7 = buffer[0]; /* +0 */

    if(RESULT_ERROR == CommandLaunch()) return RESULT_ERROR;
  }
  return(RESULT_OK);
}

/*************************************************************************
 * Function Name: FlashErase
 * Parameters:  Block Address, Block Size
 *
 * Return: 0
 *
 * Description: Erase block
 *************************************************************************/
uint32_t FlashErase(void *block_start,
                    uint32_t block_size)
{
union
{
  uint32_t  word;
  uint8_t   byte[4];
} dest;

  dest.word = (uint32_t)block_start;
  /*Set Erase command*/
  while(!(FTFA_FSTAT & FTFA_FSTAT_CCIF_MASK));

  FTFA_FCCOB0 = ERPS;
  FTFA_FCCOB1 = dest.byte[2];
  FTFA_FCCOB2 = dest.byte[1];
  FTFA_FCCOB3 = dest.byte[0];

  /*Execute Command sequence*/
  CommandLaunch();

  /* Unlock device wneh erese sector 0*/
  if(   (dest.word >= 0x400)
     && (dest.word < 0x800)
     && !user_supply_prot)
  {

    dest.word = 0x40C;
    FTFA_FCCOB0 = WRP;
    FTFA_FCCOB1 = dest.byte[2];
    FTFA_FCCOB2 = dest.byte[1];
    FTFA_FCCOB3 = dest.byte[0];
    FTFA_FCCOB4 = 0xFF; /* 0x40F */
    FTFA_FCCOB5 = 0xFF; /* 0x40E */
    FTFA_FCCOB6 = 0xFF; /* 0x40D */
    FTFA_FCCOB7 = 0xFE; /* 0x40C */

    if(RESULT_ERROR == CommandLaunch()) return RESULT_ERROR;
  }
  return(RESULT_OK);
}

/** private functions **/
static uint32_t CommandLaunch(void)
{
  /* Clear command result flags */
  FTFA_FSTAT = FTFA_FSTAT_ACCERR_MASK | FTFA_FSTAT_FPVIOL_MASK;
  /* Launch Command */
  FTFA_FSTAT = FTFA_FSTAT_CCIF_MASK;
  /* wait command end */
  while(!(FTFA_FSTAT & FTFA_FSTAT_CCIF_MASK));
  /*check for errors*/
  if(FTFA_FSTAT & (FTFA_FSTAT_ACCERR_MASK | FTFA_FSTAT_FPVIOL_MASK | FTFA_FSTAT_MGSTAT0_MASK)) return(RESULT_ERROR);
  /*No errors retur OK*/
  return (RESULT_OK);
}

static void DisableWDT(void)
{
  SIM_SCGC5    |= 1ul << 19;
  WDOG_UNLOCK   = 0xC520;
  WDOG_UNLOCK   = 0xD928;
  WDOG_STCTRLH  = 0x152;
}

